@@ -1,6 +1,6 @@
 package Font::TTF::Maxp;
 
-=head1
+=head1 NAME
 
 Font::TTF::Maxp - Maximum Profile table in a font
 
@@ -38,20 +38,35 @@ No others beyond those specified in the standard:
 =cut
 
 use strict;
-use vars qw(@ISA %fields);
+use vars qw(@ISA %fields @field_info);
 use Font::TTF::Utils;
 
 @ISA = qw(Font::TTF::Table);
+@field_info = (
+    'numGlyphs' => 'S',
+    'maxPoints' => 'S',
+    'maxContours' => 'S',
+    'maxCompositePoints' => 'S',
+    'maxCompositeContours' => 'S',
+    'maxZones' => 'S',
+    'maxTwilightPoints' => 'S',
+    'maxStorage' => 'S',
+    'maxFunctionDefs' => 'S',
+    'maxInstructionDefs' => 'S',
+    'maxStackElements' => 'S',
+    'maxSizeOfInstructions' => 'S',
+    'maxComponentElements' => 'S',
+    'maxComponentDepth' => 'S');
 
 sub init
 {
-    my ($k, $v, $c);
-    while (<Font::TTF::Maxp::DATA>)
+    my ($k, $v, $c, $i);
+    for ($i = 0; $i < $#field_info; $i += 2)
     {
-        ($k, $v, $c) = TTF_Init_Fields($_, $c);
-        next unless $k ne "";
+        ($k, $v, $c) = TTF_Init_Fields($field_info[$i], $c, $field_info[$i + 1]);
+        next unless defined $k && $k ne "";
         $fields{$k} = $v;
-    }  
+    }
 }
 
 
@@ -121,15 +136,17 @@ sub update
     my (@name) = qw(maxPoints maxContours maxCompositePoints maxCompositeContours
                     maxSizeOfInstructions maxComponentElements maxComponentDepth);
 
+    return undef unless ($self->SUPER::update);
     return undef if ($self->{'version'} == 0.5);        # only got numGlyphs
-    return undef unless (defined $self->{' PARENT'}{'loca'} && $self->{' PARENT'}{'loca'}{' isDirty'});
+    return undef unless (defined $self->{' PARENT'}{'loca'});
+    $self->{' PARENT'}{'loca'}->update;
     $num = $self->{'numGlyphs'};
 
     for ($i = 0; $i < $num; $i++)
     {
         my ($g) = $self->{' PARENT'}{'loca'}{'glyphs'}[$i] || next;
 
-        @n = $g->maxInfo($self->{' PARENT'}{'loca'}{'glyphs'});
+        @n = $g->maxInfo;
 
         for ($j = 0; $j <= $#n; $j++)
         { $m[$j] = $n[$j] if $n[$j] > $m[$j]; }
@@ -137,13 +154,12 @@ sub update
 
     foreach ('prep', 'fpgm')
     { $m[4] = length($self->{' PARENT'}{$_}{' dat'})
-            if (length($self->{' PARENT'}{$_}{' dat'}) > $m[4]);
+            if (defined $self->{' PARENT'}{$_} 
+                && length($self->{' PARENT'}{$_}{' dat'}) > $m[4]);
     }
 
     for ($j = 0; $j <= $#name; $j++)
     { $self->{$name[$j]} = $m[$j]; }
-
-    $self->{' isDirty'} = 1;
     $self;
 }
 1;
@@ -155,24 +171,7 @@ None known
 
 =head1 AUTHOR
 
-Martin Hosken L<Martin_Hosken@sil.org>. See L<Font::TTF::Font> for copyright and
+Martin Hosken Martin_Hosken@sil.org. See L<Font::TTF::Font> for copyright and
 licensing.
 
 =cut
-
-__DATA__
-numGlyphs, S
-maxPoints, S
-maxContours, S
-maxCompositePoints, S
-maxCompositeContours, S
-maxZones, S
-maxTwilightPoints, S
-maxStorage, S
-maxFunctionDefs, S
-maxInstructionDefs, S
-maxStackElements, S
-maxSizeOfInstructions, S
-maxComponentElements, S
-maxComponentDepth, S
-
