@@ -78,8 +78,7 @@ This holds the index of the default feature, if there is one, or -1 otherwise.
 
 =item FEATURES
 
-This is an array of feature indices which index into the FEATURES instance
-variable of the table
+This is an array of feature tags for all the features enabled for this language
 
 =back
 
@@ -323,7 +322,7 @@ sub read
 	    push (@{$l->{'FEAT_TAGS'}}, $tag);
     }
 
-    foreach $tag (grep {length($_) == 4} keys %$l)
+    foreach $tag (grep {m/^.{4}(?:\s_\d+)?$/o} keys %$l)
     {
 	    $fh->seek($moff + $l->{$tag}{' OFFSET'}, 0);
     	$fh->read($dat, 4);
@@ -364,7 +363,7 @@ sub read
     	    ($lTag, $off) = unpack("a4n", substr($dat, $i * 6, 6));
     	    $off += $l->{$tag}{' OFFSET'};
     	    $l->{$tag}{$lTag}{' OFFSET'} = $off;
-            foreach (@{$l->{$tag}{'LANG_TAGS'}})
+            foreach (@{$l->{$tag}{'LANG_TAGS'}}, 'DEFAULT')
             { $l->{$tag}{$lTag}{' REFTAG'} = $_ if ($l->{$tag}{$_}{' OFFSET'} == $off
                                                    && !$l->{$tag}{$_}{' REFTAG'}); }
     	    push (@{$l->{$tag}{'LANG_TAGS'}}, $lTag);
@@ -413,10 +412,20 @@ sub read
     	($l->{'TYPE'}, $l->{'FLAG'}, $nSub) = unpack("n3", $dat);
     	$fh->read($dat, $nSub * 2);
     	$j = 0;
-    	map { $l->{'SUB'}[$j]{' OFFSET'} = $_; } unpack("n*", $dat);
+        my @offsets = unpack("n*", $dat);
+        my $isExtension = ($l->{'TYPE'} == $self->extension());
     	for ($j = 0; $j < $nSub; $j++)
     	{
-    	    $fh->seek($moff + $oLook + $l->{' OFFSET'} + $l->{'SUB'}[$j]{' OFFSET'}, 0);
+    	    $l->{'SUB'}[$j]{' OFFSET'} = $offsets[$j];
+       	    $fh->seek($moff + $oLook + $l->{' OFFSET'} + $l->{'SUB'}[$j]{' OFFSET'}, 0);
+    	    if ($isExtension)
+    	    {
+    	        $fh->read($dat, 8);
+    	        my $longOff;
+    	        (undef, $l->{'TYPE'}, $longOff) = unpack("nnN", $dat);
+    	        $l->{'SUB'}[$j]{' OFFSET'} += $longOff;
+        	    $fh->seek($moff + $oLook + $l->{' OFFSET'} + $l->{'SUB'}[$j]{' OFFSET'}, 0);
+            }
 	        $self->read_sub($fh, $l, $j);
 	    }
     }
@@ -460,13 +469,13 @@ sub out
 {
     my ($self, $fh) = @_;
     my ($i, $j, $base, $off, $tag, $t, $l, $lTag, $oScript, @script, @tags);
-    my ($end, $nTags, @offs, $oFeat, $oLook, $nSub, $nSubs, $big);
+    my ($end, $nTags, @offs, $oFeat, $oLook, $nSub, $nSubs, $big, $out);
 
     return $self->SUPER::out($fh) unless $self->{' read'};
 
 # First sort the features
     $i = 0;
-    $self->{'FEATURES'}{'FEAT_TAGS'} = [sort grep {length($_) == 4 || m/\s_\d+$/o} %{$self->{'FEATURES'}}]
+    $self->{'FEATURES'}{'FEAT_TAGS'} = [sort grep {m/^.{4}(?:\s_\d+)?$/o} %{$self->{'FEATURES'}}]
             if (!defined $self->{'FEATURES'}{'FEAT_TAGS'});
     foreach $t (@{$self->{'FEATURES'}{'FEAT_TAGS'}})
     { $self->{'FEATURES'}{$t}{'INDEX'} = $i++; }
@@ -494,15 +503,15 @@ sub out
     	{
     	    my ($def);
     	    $l = $tag->{$lTag};
-    	    next if (!defined $l || $l->{' REFTAG'} ne '');
+    	    next if (!defined $l || (defined $l->{' REFTAG'} && $l->{' REFTAG'} ne ''));
     	    $l->{' OFFSET'} = tell($fh) - $base - $oScript - $tag->{' OFFSET'};
     	    if (defined $l->{'DEFAULT'})
 #    	    { $def = $self->{'FEATURES'}{$l->{'FEATURES'}[$l->{'DEFAULT'}]}{'INDEX'}; }
             { $def = $l->{'DEFAULT'}; }
     	    else
     	    { $def = -1; }
-    	    $fh->print(pack("n*", $l->{'RE_ORDER'}, $def, $#{$l->{'FEATURES'}} + 1,
-    	            map {$self->{'FEATURES'}{$_}{'INDEX'}} @{$l->{'FEATURES'}}));
+    	    $fh->print(pack("n*", $l->{'RE_ORDER'} || 0, $def, $#{$l->{'FEATURES'}} + 1,
+    	            map {$self->{'FEATURES'}{$_}{'INDEX'} || 0} @{$l->{'FEATURES'}}));
     	}
     	$end = $fh->tell();
     	if ($tag->{'DEFAULT'}{' REFTAG'} || defined $tag->{'DEFAULT'}{'FEATURES'})
@@ -600,12 +609,16 @@ sub out
         }
         else
         { $end = $tag->{' EXT_OFFSET'}; }
-        @offs = ();
+        my (@offs, $out, @refs);
         for ($j = 0; $j < $nSub; $j++)
         {
-            push(@offs, tell($fh) - $end);
-            $self->out_sub($fh, $tag, $j);
+            my ($ctables) = {};
+            my ($base) = length($out);
+            push(@offs, tell($fh) - $end + $base);
+            $out .= $self->out_sub($fh, $tag, $j, $ctables, $base);
+            push (@refs, [$ctables, $base]);
         }
+        out_final($fh, $out, \@refs);
         $end = $fh->tell();
         if (!defined $big)
         {
@@ -745,11 +758,11 @@ by storing in a string first and the return value is "";
 sub out_final
 {
     my ($fh, $out, $cache_list, $state) = @_;
-    my ($len) = length($out);
+    my ($len) = length($out || '');
     my ($base_loc) = $state ? 0 : $fh->tell();
-    my ($loc, $t, $r, $s, $master_cache, $offs, $str);
+    my ($loc, $t, $r, $s, $master_cache, $offs, $str, %vecs);
 
-    $fh->print($out) unless $state;       # first output the current attempt
+    $fh->print($out || '') unless $state;       # first output the current attempt
     foreach $r (@$cache_list)
     {
         $offs = $r->[1];
@@ -758,12 +771,19 @@ sub out_final
             $str = "$t";
             if (!defined $master_cache->{$str})
             {
-                $master_cache->{$str} = ($state ? length($out) : $fh->tell())
-                                                            - $base_loc;
-                if ($state)
-                { $out .= $r->[0]{$str}[0]->out($fh, 1); }
+                my ($vec) = $r->[0]{$str}[0]->signature();
+                if ($vecs{$vec})
+                { $master_cache->{$str} = $master_cache->{$vecs{$vec}}; }
                 else
-                { $r->[0]{$str}[0]->out($fh, 0); }
+                {
+                    $vecs{$vec} = $str;
+                    $master_cache->{$str} = ($state ? length($out) : $fh->tell())
+                                                                       - $base_loc;
+                    if ($state)
+                    { $out .= $r->[0]{$str}[0]->out($fh, 1); }
+                    else
+                    { $r->[0]{$str}[0]->out($fh, 0); }
+                }
             }
             foreach $s (@{$r->[0]{$str}[1]})
             { substr($out, $s, 2) = pack('n', $master_cache->{$str} - $offs); }
@@ -775,7 +795,7 @@ sub out_final
     {
         $loc = $fh->tell();
         $fh->seek($base_loc, 0);
-        $fh->print($out);       # the corrected version
+        $fh->print($out || '');       # the corrected version
         $fh->seek($loc, 0);
     }
 }
@@ -964,29 +984,30 @@ in a GSUB table so calling from GPOS should adjust the value accordingly.
 
 sub out_context
 {
-    my ($self, $lookup, $fh, $type, $fmt, $ctables, $out, $num) = @_;
+    my ($self, $lookup, $fh, $type, $fmt, $ctables, $out, $num, $base) = @_;
     my ($offc, $offd, $i, $j, $r, $t, $numd);
 
+    $out ||= '';
     if (($type == 4 || $type == 5 || $type == 6) && ($fmt == 1 || $fmt == 2))
     {
         my ($base_off);
         
         if ($fmt == 1)
         {
-            $out = pack("nnn", $fmt, Font::TTF::Ttopen::ref_cache($lookup->{'COVERAGE'}, $ctables, 2),
+            $out = pack("nnn", $fmt, Font::TTF::Ttopen::ref_cache($lookup->{'COVERAGE'}, $ctables, 2 + $base),
                             $num);
             $base_off = 6;
         } elsif ($type == 5)
         {
-            $out = pack("nnnn", $fmt, Font::TTF::Ttopen::ref_cache($lookup->{'COVERAGE'}, $ctables, 2),
-                            Font::TTF::Ttopen::ref_cache($lookup->{'CLASS'}, $ctables, 4), $num);
+            $out = pack("nnnn", $fmt, Font::TTF::Ttopen::ref_cache($lookup->{'COVERAGE'}, $ctables, 2 + $base),
+                            Font::TTF::Ttopen::ref_cache($lookup->{'CLASS'}, $ctables, 4 + $base), $num);
             $base_off = 8;
         } elsif ($type == 6)
         {
-            $out = pack("n6", $fmt, Font::TTF::Ttopen::ref_cache($lookup->{'COVERAGE'}, $ctables, 2),
-                                Font::TTF::Ttopen::ref_cache($lookup->{'PRE_CLASS'}, $ctables, 4),
-                                Font::TTF::Ttopen::ref_cache($lookup->{'CLASS'}, $ctables, 6),
-                                Font::TTF::Ttopen::ref_cache($lookup->{'POST_CLASS'}, $ctables, 8),
+            $out = pack("n6", $fmt, Font::TTF::Ttopen::ref_cache($lookup->{'COVERAGE'}, $ctables, 2 + $base),
+                                Font::TTF::Ttopen::ref_cache($lookup->{'PRE_CLASS'}, $ctables, 4 + $base),
+                                Font::TTF::Ttopen::ref_cache($lookup->{'CLASS'}, $ctables, 6 + $base),
+                                Font::TTF::Ttopen::ref_cache($lookup->{'POST_CLASS'}, $ctables, 8 + $base),
                                 $num);
             $base_off = 12;
         }
@@ -1033,23 +1054,23 @@ sub out_context
         $out .= pack('n3', $fmt, $#{$lookup->{'RULES'}[0][0]{'MATCH'}} + 1,
                                 $#{$lookup->{'RULES'}[0][0]{'ACTION'}} + 1);
         foreach $t (@{$lookup->{'RULES'}[0][0]{'MATCH'}})
-        { $out .= pack('n', Font::TTF::Ttopen::ref_cache($t, $ctables, length($out))); }
+        { $out .= pack('n', Font::TTF::Ttopen::ref_cache($t, $ctables, length($out) + $base)); }
         foreach $t (@{$lookup->{'RULES'}[0][0]{'ACTION'}})
         { $out .= pack('n2', @$t); }
     } elsif ($type == 6 && $fmt == 3)
     {
         $r = $lookup->{'RULES'}[0][0];
 		no strict 'refs';	# temp fix - more code needed (probably "if" statements in the event 'PRE' or 'POST' are empty)
-        $out .= pack('n2', $fmt, scalar @{$r->{'PRE'}});
+        $out .= pack('n2', $fmt, defined $r->{'PRE'} ? scalar @{$r->{'PRE'}} : 0);
         foreach $t (@{$r->{'PRE'}})
-        { $out .= pack('n', Font::TTF::Ttopen::ref_cache($t, $ctables, length($out))); }
-        $out .= pack('n', scalar @{$r->{'MATCH'}});
+        { $out .= pack('n', Font::TTF::Ttopen::ref_cache($t, $ctables, length($out) + $base)); }
+        $out .= pack('n', defined $r->{'MATCH'} ? scalar @{$r->{'MATCH'}} : 0);
         foreach $t (@{$r->{'MATCH'}})
-        { $out .= pack('n', Font::TTF::Ttopen::ref_cache($t, $ctables, length($out))); }
-        $out .= pack('n', scalar @{$r->{'POST'}});
+        { $out .= pack('n', Font::TTF::Ttopen::ref_cache($t, $ctables, length($out) + $base)); }
+        $out .= pack('n', defined $r->{'POST'} ? scalar @{$r->{'POST'}} : 0);
         foreach $t (@{$r->{'POST'}})
-        { $out .= pack('n', Font::TTF::Ttopen::ref_cache($t, $ctables, length($out))); }
-        $out .= pack('n', scalar @{$r->{'ACTION'}});
+        { $out .= pack('n', Font::TTF::Ttopen::ref_cache($t, $ctables, length($out) + $base)); }
+        $out .= pack('n', defined $r->{'ACTION'} ? scalar @{$r->{'ACTION'}} : 0);
         foreach $t (@{$r->{'ACTION'}})
         { $out .= pack('n2', @$t); }
     }
